# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Defines fetching data from Bid Manager API."""

from collections.abc import Sequence

import garf_bid_manager
import pydantic
from garf_core import report

from media_fetching.sources import models


class BidManagerFetchingParameters(models.FetchingParameters):
  """YouTube specific parameters for getting media data."""

  advertiser: str
  metrics: Sequence[str] = [
    'clicks',
    'impressions',
  ]
  segments: Sequence[str] | None = pydantic.Field(default_factory=list)
  extra_info: Sequence[str] | None = pydantic.Field(default_factory=list)


class Fetcher(models.BaseMediaInfoFetcher):
  """Extracts media information from Bid Manager API."""

  def fetch_media_data(
    self,
    fetching_request: BidManagerFetchingParameters,
  ) -> report.GarfReport:
    """Fetches performance data from Bid Manager API."""
    fetcher = garf_bid_manager.BidManagerApiReportFetcher()
    query = """
      SELECT
        date AS date,
        youtube_ad_video_id AS media_url,
        youtube_ad_video AS media_name,
        metric_impressions AS impressions,
        metric_clicks AS clicks
      FROM youtube
      WHERE advertiser = {advertiser}
      AND dataRange = LAST_30_DAYS
    """
    return fetcher.fetch(query.format(advertiser=fetching_request.advertiser))



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Fetches media data from various sources."""

# pylint: disable=C0330, g-bad-import-order, g-multiple-import

from media_fetching.media_fetching_service import MediaFetchingService
from media_fetching.sources.dbm import BidManagerFetchingParameters
from media_fetching.sources.fake import FakeFetchingParameters
from media_fetching.sources.file import FileFetchingParameters
from media_fetching.sources.googleads import GoogleAdsFetchingParameters
from media_fetching.sources.sql import (
  BigQueryFetchingParameters,
  SqlAlchemyQueryFetchingParameters,
)
from media_fetching.sources.youtube import YouTubeFetchingParameters

__all__ = [
  'MediaFetchingService',
]

INPUT_MAPPING = {
  'googleads': GoogleAdsFetchingParameters,
  'youtube': YouTubeFetchingParameters,
  'file': FileFetchingParameters,
  'sqldb': SqlAlchemyQueryFetchingParameters,
  'bq': BigQueryFetchingParameters,
  'fake': FakeFetchingParameters,
  'dbm': BidManagerFetchingParameters,
}
__version__ = '0.3.4'



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Provides HTTP endpoint for filonov requests."""

# pylint: disable=C0330, g-bad-import-order, g-multiple-import

import fastapi
import pydantic
import uvicorn
from garf_io import writer

import media_fetching
from media_fetching.sources import models

router = fastapi.APIRouter(prefix='/media_fetching')


class WriterOptions(pydantic.BaseModel):
  writer: str = 'json'
  writer_parameters: dict[str, str] = pydantic.Field(default_factory=dict)
  output: str = 'media_results'


@router.post('/fetch:file')
async def fetch_file(
  request: media_fetching.sources.file.FileFetchingParameters,
  writer_options: WriterOptions,
) -> fastapi.responses.JSONResponse:
  """Fetches media data from a file."""
  return fetch('file', request, writer_options)


@router.post('/fetch:googleads')
async def fetch_googleads(
  request: media_fetching.sources.googleads.GoogleAdsFetchingParameters,
  writer_options: WriterOptions,
) -> fastapi.responses.JSONResponse:
  """Fetches media data from Google Ads."""
  return fetch('googleads', request, writer_options)


@router.post('/fetch:youtube')
async def fetch_youtube(
  request: media_fetching.sources.youtube.YouTubeFetchingParameters,
  writer_options: WriterOptions,
) -> fastapi.responses.JSONResponse:
  """Fetches media data from YouTube."""
  return fetch('youtube', request, writer_options)


@router.post('/fetch:bq')
async def fetch_bq(
  request: media_fetching.sources.sql.BigQueryFetchingParameters,
  writer_options: WriterOptions,
) -> fastapi.responses.JSONResponse:
  """Fetches media data from BigQuery."""
  return fetch('bq', request, writer_options)


@router.post('/fetch:sqldb')
async def fetch_sqldb(
  request: media_fetching.sources.sql.SqlAlchemyQueryFetchingParameters,
  writer_options: WriterOptions,
) -> fastapi.responses.JSONResponse:
  """Fetches media data from SqlAlchemy DB."""
  return fetch('sqldb', request, writer_options)


@router.post('/fetch:dbm')
async def fetch_dbm(
  request: media_fetching.sources.dbm.BidManagerFetchingParameters,
  writer_options: WriterOptions,
) -> fastapi.responses.JSONResponse:
  """Fetches media data from Bid Manager API."""
  return fetch('dbm', request, writer_options)


def fetch(
  source: str | models.InputSource,
  request: models.FetchingParameters,
  writer_options: WriterOptions,
):
  """Fetches media data from a provided source."""
  fetching_service = media_fetching.MediaFetchingService.from_source_alias(
    source
  )
  report = fetching_service.fetch(request)
  return writer.create_writer(
    writer_options.writer, **writer_options.writer_parameters
  ).write(report, writer_options.output)


app = fastapi.FastAPI()
app.include_router(router)


def main():
  uvicorn.run(app)


if __name__ == '__main__':
  main()



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=C0330, g-bad-import-order, g-multiple-import

"""Defines imports from Google Ads Reports."""

import datetime
import functools
import operator
import os
import pathlib
from collections.abc import Mapping, Sequence
from typing import Literal, get_args

import gaarf
import garf_youtube_data_api
import pydantic
from garf_core import report
from media_tagging import media

from media_fetching import exceptions
from media_fetching.sources import models, queries


class GoogleAdsFetchingParameters(models.FetchingParameters):
  """Google Ads specific parameters for getting media data."""

  account: str
  media_type: Literal[tuple(media.MediaTypeEnum.options())] | None = None
  start_date: str = (
    datetime.datetime.today() - datetime.timedelta(days=30)
  ).strftime('%Y-%m-%d')
  end_date: str = (
    datetime.datetime.today() - datetime.timedelta(days=1)
  ).strftime('%Y-%m-%d')
  min_cost: int = 0
  campaign_types: Sequence[queries.SupportedCampaignTypes] | str = ('app',)
  ads_config: str = os.getenv(
    'GOOGLE_ADS_CONFIGURATION_FILE_PATH',
    str(pathlib.Path.home() / 'google-ads.yaml'),
  )
  metrics: Sequence[str] | str = [
    'clicks',
    'impressions',
    'cost',
    'conversions',
    'conversions_value',
  ]
  segments: Sequence[str] | str = ('format_type', 'channel_type')
  extra_info: str | Sequence[str] | None = pydantic.Field(default_factory=list)

  def model_post_init(self, __context__):  # noqa: D105
    if self.campaign_types == 'all':
      self.campaign_types = get_args(queries.SupportedCampaignTypes)
    elif isinstance(self.campaign_types, str):
      self.campaign_types = self.campaign_types.split(',')
    if isinstance(self.segments, str):
      self.segments = self.segments.split(',')
    if isinstance(self.metrics, str):
      self.metrics = self.metrics.split(',')
    if isinstance(self.extra_info, str):
      self.extra_info = self.extra_info.split(',')

  @property
  def query_params(self) -> dict[str, str | int]:
    query_params = {
      'start_date': self.start_date,
      'end_date': self.end_date,
      'media_type': self.media_type,
      'min_cost': self.min_cost,
    }
    query_params.update(self.model_extra)
    return query_params


class Fetcher(models.BaseMediaInfoFetcher):
  """Extracts media information from Google Ads."""

  def fetch_media_data(
    self,
    fetching_request: GoogleAdsFetchingParameters,
  ) -> report.GarfReport:
    """Fetches performance data from Google Ads API.

    Args:
      fetching_request: Request for getting data from Google Ads.

    Returns:
      Report with performance data.

    Raises:
      MediaFetchingError: When there's no data for a given fetching context.

    """
    if isinstance(fetching_request, Mapping):
      fetching_request = GoogleAdsFetchingParameters(**fetching_request)
    self.fetcher = gaarf.AdsReportFetcher(
      api_client=gaarf.GoogleAdsApiClient(
        path_to_config=fetching_request.ads_config
      )
    )

    performance_queries = self._define_performance_queries(fetching_request)
    self.accounts = self._define_customer_ids(fetching_request)
    performance = self._execute_performance_queries(
      performance_queries=performance_queries,
      fetching_request=fetching_request,
    )
    if not performance:
      raise exceptions.MediaFetchingError(
        f'No performance data found for the context: {fetching_request}'
      )
    self._add_in_campaigns(performance)
    if fetching_request.media_type == 'YOUTUBE_VIDEO':
      self._add_video_info(performance)

    return performance

  def _define_performance_queries(
    self, fetching_request: GoogleAdsFetchingParameters
  ) -> dict[str, queries.PerformanceQuery]:
    """Defines queries based on campaign and media types.

    Args:
      fetching_request: Request for fetching data from Google Ads.

    Returns:
      Mapping between each campaign type and its corresponding query.
    """
    performance_queries = {}
    for campaign_type in fetching_request.campaign_types:
      query = queries.QUERIES_MAPPING.get(campaign_type)
      if (
        campaign_type == 'video' and fetching_request.media_type == 'IMAGE'
      ) or (
        campaign_type == 'search' and fetching_request.media_type != 'TEXT'
      ):
        continue
      if campaign_type == 'demandgen':
        query = query.get(fetching_request.media_type)
      performance_queries[campaign_type] = query
    return performance_queries

  def _define_customer_ids(
    self,
    fetching_request: GoogleAdsFetchingParameters,
  ) -> list[str]:
    """Identifies all accounts that have campaigns with specified types.

    Args:
      fetching_request: Request for fetching data from Google Ads.

    Returns:
      All accounts that have campaigns with specified types.
    """
    campaign_types = ','.join(
      queries.CAMPAIGN_TYPES_MAPPING.get(campaign_type)
      for campaign_type in fetching_request.campaign_types
    )
    customer_ids_query = (
      'SELECT customer.id FROM campaign '
      f'WHERE campaign.advertising_channel_type IN ({campaign_types})'
    )
    return self.fetcher.expand_mcc(fetching_request.account, customer_ids_query)

  def _execute_performance_queries(
    self,
    performance_queries: dict[str, queries.PerformanceQuery],
    fetching_request: GoogleAdsFetchingParameters,
  ) -> report.GarfReport:
    """Executes performance queries for a set of customer ids.

    If two or more performance queries are specified only common fields are
    included into the resulting report.

    Args:
      performance_queries: Queries that need to be executed.
      fetching_request: Request for fetching data from Google Ads.

    Returns:
      Report with media performance.
    """
    performance_reports = []
    for campaign_type, query in performance_queries.items():
      fetching_parameters = fetching_request.query_params
      fetching_parameters['campaign_type'] = campaign_type
      performance = self.fetcher.fetch(
        query(**fetching_parameters),
        self.accounts,
      )
      if len(performance_queries) == 1:
        return performance
      if performance:
        performance_reports.append(performance)
    common_fields = []
    for performance_report in performance_reports:
      common_fields.append(set(performance_report.column_names))
    common_fields = list(set.intersection(*common_fields))
    performance_reports = [
      performance_report[common_fields]
      for performance_report in performance_reports
    ]
    return functools.reduce(operator.add, performance_reports)

  def _add_in_campaigns(
    self,
    performance: report.GarfReport,
  ) -> None:
    """Injects number of campaigns media is presented in into report."""
    info = {
      media_url: str(len(campaigns))
      for media_url, campaigns in performance.to_dict(
        'media_url', 'campaign_id'
      ).items()
    }
    for row in performance:
      row['in_campaigns'] = info.get(row.media_url)

  def _add_video_info(
    self,
    performance: report.GarfReport,
  ) -> None:
    """Injects YouTube specific information on media.

    Args:
      performance: Report to add video data into.

    Returns:
      Mapping between video id and its information.
    """
    video_durations_query = """
     SELECT
       media_file.video.youtube_video_id AS video_id,
       media_file.video.ad_duration_millis / 1000 AS video_duration
     FROM media_file
     WHERE media_file.type = VIDEO
    """
    video_orientations_query = """
    SELECT
      id,
      player.embedWidth AS width,
      player.embedHeight AS height
    FROM videos
    """

    video_durations = {
      video_id: video_lengths[0]
      for video_id, video_lengths in self.fetcher.fetch(
        video_durations_query, self.accounts
      )
      .to_dict(
        key_column='video_id',
        value_column='video_duration',
      )
      .items()
    }

    video_ids = performance['media_url'].to_list(
      row_type='scalar', distinct=True
    )
    youtube_api_fetcher = garf_youtube_data_api.YouTubeDataApiReportFetcher()
    video_orientations = youtube_api_fetcher.fetch(
      video_orientations_query,
      id=video_ids,
      maxWidth=500,
    )

    for row in video_orientations:
      aspect_ratio = round(int(row.width) / int(row.height), 2)
      if aspect_ratio > 1:
        row['orientation'] = 'Landscape'
      elif aspect_ratio < 1:
        row['orientation'] = 'Portrait'
      else:
        row['orientation'] = 'Square'

    video_orientations = video_orientations.to_dict(
      key_column='id',
      value_column='orientation',
      value_column_output='scalar',
    )
    for row in performance:
      video_id = row.media_url
      row['orientation'] = video_orientations.get(video_id, 0.0)
      row['duration'] = video_durations.get(video_id, 0.0)



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=C0330, g-bad-import-order, g-multiple-import

"""Defines fetching data from a DB."""

from collections.abc import Sequence
from typing import Literal

import pydantic
from garf_core import report
from garf_executors import bq_executor, sql_executor

from media_fetching.sources import models


class TableFetchingParameters(models.FetchingParameters):
  """SQL table specific parameters for getting media."""

  table: str
  media_type: Literal['IMAGE', 'VIDEO', 'YOUTUBE_VIDEO', None] = None
  media_identifier: str = 'media_url'
  media_name: str = 'media_name'
  metrics: Sequence[str] | str = ('clicks', 'impressions')
  segments: Sequence[str] | str | None = None
  extra_info: Sequence[str] | None = pydantic.Field(default_factory=list)

  @property
  def query_text(self) -> str:
    fields = f'{self.media_identifier}, {self.media_name}'
    if self.metrics:
      metrics = ', '.join(self.metrics)
      if metrics:
        fields = f'{fields}, {metrics}'
    if self.segments:
      segments = ', '.join(self.segments)
      if segments:
        fields = f'{fields}, {segments}'
    return f'SELECT {fields} FROM {self.table}'

  def model_post_init(self, __context__):
    if isinstance(self.metrics, str):
      self.metrics = self.metrics.split(',')
    if isinstance(self.segments, str):
      self.segments = self.segments.split(',')


class BigQueryFetchingParameters(TableFetchingParameters):
  """BigQuery specific parameters for getting media."""

  @property
  def project(self) -> str:
    return self.table.split('.')[0]


class SqlAlchemyQueryFetchingParameters(TableFetchingParameters):
  """SqlAlchemy specific parameters for getting media."""

  connection_string: str = pydantic.Field(exclude=True)


class BigQueryFetcher(models.BaseMediaInfoFetcher):
  """Extracts information from a BigQuery table into a media report."""

  def fetch_media_data(
    self,
    fetching_request: BigQueryFetchingParameters,
  ) -> report.GarfReport:
    executor = bq_executor.BigQueryExecutor(
      project_id=fetching_request.project,
    )
    return executor.execute(
      title='media_data', query=fetching_request.query_text
    )


class SqlAlchemyQueryFetcher(models.BaseMediaInfoFetcher):
  """Extracts information from a SQL database into a media report."""

  def fetch_media_data(
    self,
    fetching_request: SqlAlchemyQueryFetchingParameters,
  ) -> report.GarfReport:
    executor = sql_executor.SqlAlchemyQueryExecutor.from_connection_string(
      connection_string=fetching_request.connection_string
    )
    return executor.execute(
      title='media_data', query=fetching_request.query_text
    )



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=C0330, g-bad-import-order, g-multiple-import

"""Defines fetching data from a file."""

import os
from collections.abc import Sequence
from typing import Literal

import pandas as pd
import pydantic
import smart_open
from garf_core import report
from media_tagging import media

from media_fetching.sources import models


class FileFetchingParameters(models.FetchingParameters):
  """File specific parameters for getting media."""

  path: os.PathLike[str] | str
  media_type: Literal[tuple(media.MediaTypeEnum.options())] | None = None
  media_identifier: str = 'media_url'
  media_name: str = 'media_name'
  metrics: Sequence[str] | str = ('clicks', 'impressions')
  segments: Sequence[str] | None = None
  extra_info: Sequence[str] | None = pydantic.Field(default_factory=list)

  def model_post_init(self, __context__):
    if isinstance(self.metrics, str):
      self.metrics = self.metrics.split(',')


class Fetcher(models.BaseMediaInfoFetcher):
  """Extracts information from a file to a media report."""

  def fetch_media_data(
    self,
    fetching_request: FileFetchingParameters,
  ) -> report.GarfReport:
    return report.GarfReport.from_pandas(
      pd.read_csv(smart_open.open(fetching_request.path))
    )



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=C0330, g-bad-import-order, g-multiple-import

"""Google Ads specific enrichers."""

import functools
import os
import pathlib

import gaarf
from garf_core import report

from media_fetching import exceptions
from media_fetching.enrichers import extra_info


class GoogleAdsEnricherError(exceptions.MediaFetchingError):
  """Google Ads specific exceptions for enricher."""


class GoogleAdsEnricher:
  """Injects Google Ads specific information into existing reports."""

  def __init__(
    self, account: str, ads_config: str | None = None, **kwargs: str
  ) -> None:
    """Initializes MediaTaggingEnricher."""
    self.account = account
    self.ads_config = ads_config or os.getenv(
      'GOOGLE_ADS_CONFIGURATION_FILE_PATH',
      str(pathlib.Path.home() / 'google-ads.yaml'),
    )
    self.kwargs = kwargs

  @functools.cached_property
  def fetcher(self) -> gaarf.AdsReportFetcher:
    return gaarf.AdsReportFetcher(
      api_client=gaarf.GoogleAdsApiClient(path_to_config=self.ads_config)
    )

  @functools.cached_property
  def accounts(self) -> list[str]:
    return self.fetcher.expand_mcc(self.account)

  def approval_rate(
    self, performance: report.GarfReport, **kwargs: str
  ) -> extra_info.ExtraInfo:
    """Calculates percentage of approvals for each media.

    Args:
      performance: Report with performance data.

    Returns:
      Mapping between media_url and approval rate (from 0 to 1).
    """
    approvals = performance.to_dict(
      key_column='media_url', value_column='approval_status'
    )
    approval_rates = {}
    for media_url, statuses in approvals.items():
      approval_rate = 1 - len(
        list(filter(lambda x: x != 'APPROVED', statuses))
      ) / len(statuses)

      approval_rates[media_url] = {'approval_rate': approval_rate}
    return extra_info.ExtraInfo(info=approval_rates)

  def main_geo(
    self, performance: report.GarfReport, **kwargs: str
  ) -> extra_info.ExtraInfo:
    """Calculates dominant country for each media.

    Dominant country is calculated on campaign level based on >50% of all
    spend and propagated to all media in this campaign.

    Args:
      performance: Report with performance data.

    Returns:
      Mapping between media_url and its dominant country.

    Raises:
      GoogleAdsEnricherError: If report does not have campaign_id column.
    """
    if 'campaign_id' not in performance.column_names:
      raise GoogleAdsEnricherError(
        '"campaign_id" is required for "main_geo" enriching'
      )
    threshold = 0.5

    def get_dominant_country(group):
      dominant_country_row = group[group['share'] > threshold]
      if dominant_country_row.empty:
        return 'Unknown'
      return dominant_country_row['country'].iloc[0]

    campaign_geos = """
    SELECT
      campaign.id AS campaign_id,
      user_location_view.country_criterion_id AS country_id,
     metrics.cost_micros / 1e6 AS cost
    FROM user_location_view
    """

    geo_targets_query = """
    SELECT
      geo_target_constant.id AS country_id,
      geo_target_constant.name AS country_name
    FROM geo_target_constant
    WHERE geo_target_constant.id BETWEEN 2000 AND 3000
    """

    country_mapping = self.fetcher.fetch(
      geo_targets_query, self.accounts[0]
    ).to_dict(
      key_column='country_id',
      value_column='country_name',
      value_column_output='scalar',
    )
    geo_extra_info = self.fetcher.fetch(
      campaign_geos, self.accounts
    ).to_pandas()
    geo_extra_info['country'] = geo_extra_info['country_id'].map(
      country_mapping
    )
    geo_extra_info['country'] = geo_extra_info['country'].fillna('Unknown')
    geo_extra_info['total_campaign_cost'] = geo_extra_info.groupby(
      'campaign_id'
    )['cost'].transform('sum')
    geo_extra_info['share'] = (
      geo_extra_info['cost'] / geo_extra_info['total_campaign_cost']
    )
    geo_info = (
      geo_extra_info.groupby('campaign_id')
      .apply(get_dominant_country)
      .to_dict()
    )
    return extra_info.ExtraInfo(
      info={
        campaign_id: {'main_geo': value}
        for campaign_id, value in geo_info.items()
      },
      base_key='campaign_id',
    )



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=C0330, g-bad-import-order, g-multiple-import

"""Media tagging specific enrichers."""

from typing import Literal

import media_tagging
from garf_core import report

from media_fetching.enrichers import extra_info


class MediaTaggingEnricher:
  """Uses media tagging library to add extra information into reports."""

  def __init__(self, db_uri: str | None = None) -> None:
    """Initializes MediaTaggingEnricher."""
    self.tagging_service = media_tagging.MediaTaggingService(
      media_tagging.repositories.SqlAlchemyTaggingResultsRepository(db_uri)
    )

  def language(
    self,
    performance: report.GarfReport,
    media_type: Literal['IMAGE', 'YOUTUBE_VIDEO'],
  ) -> extra_info.ExtraInfo:
    """Infers languages of each media."""
    media_urls = performance['media_url'].to_list(
      row_type='scalar', distinct=True
    )
    media_path_mapping = {
      media_tagging.media.convert_path_to_media_name(media_url): media_url
      for media_url in media_urls
    }
    languages = self.tagging_service.describe_media(
      media_tagging.MediaTaggingRequest(
        tagger_type='gemini',
        media_type=media_type,
        media_paths=media_urls,
        tagging_options={
          'custom_prompt': (
            f'What is the language of this {media_type}? '
            'Provide only a lowercase name of the language '
            '(english, spanish, japanese).'
          )
        },
      )
    )
    return extra_info.ExtraInfo(
      info={
        media_path_mapping.get(r.identifier): {'language': r.content.text}
        for r in languages.results
      },
    )



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=C0330, g-bad-import-order, g-multiple-import

"""YouTube specific enrichers."""

import garf_youtube_data_api
from garf_core import report

from media_fetching import exceptions
from media_fetching.enrichers import extra_info


class YouTubeEnricherError(exceptions.MediaFetchingError):
  """Google Ads specific exceptions for enricher."""


class YouTubeEnricher:
  """Injects YouTube specific information into existing reports."""

  @property
  def fetcher(self):
    """Initialized Report fetcher from YouTube Data API."""
    return garf_youtube_data_api.YouTubeDataApiReportFetcher()

  def language(
    self, performance: report.GarfReport, **kwargs: str
  ) -> extra_info.ExtraInfo:
    """Gets either visual or spoken language of YouTube video."""
    video_ids = performance['media_url'].to_list(
      row_type='scalar', distinct=True
    )
    language_query = """
    SELECT
      id,
      snippet.defaultLanguage AS language,
      snippet.defaultAudioLanguage AS audio_language
    FROM videos
    """
    languages = self.fetcher.fetch(
      language_query,
      id=video_ids,
    )
    return extra_info.ExtraInfo(
      info={
        r.id: {'language': r.language or r.audio_language or 'Unknown'}
        for r in languages
      },
    )



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=C0330, g-bad-import-order, g-multiple-import

"""Defines fetching data from YouTube channel."""

from collections.abc import Sequence

import garf_youtube_data_api
import pydantic
from garf_core import report

from media_fetching.sources import models


class YouTubeFetchingParameters(models.FetchingParameters):
  """YouTube specific parameters for getting media data."""

  channel: str
  metrics: Sequence[str] = [
    'views',
    'likes',
  ]
  segments: Sequence[str] | None = None
  extra_info: Sequence[str] | None = pydantic.Field(default_factory=list)


class Fetcher(models.BaseMediaInfoFetcher):
  """Extracts media information from YouTube Data API."""

  def fetch_media_data(
    self,
    fetching_request: YouTubeFetchingParameters,
  ) -> report.GarfReport:
    """Get all public videos from YouTube channel."""
    youtube_api_fetcher = garf_youtube_data_api.YouTubeDataApiReportFetcher()
    channel_uploads_playlist_query = """
    SELECT
      contentDetails.relatedPlaylists.uploads AS uploads_playlist
    FROM channels
    """
    videos_playlist = youtube_api_fetcher.fetch(
      channel_uploads_playlist_query,
      id=[fetching_request.channel],
    )

    channel_videos_query = """
    SELECT
      contentDetails.videoId AS video_id
    FROM playlistItems
    """
    videos = youtube_api_fetcher.fetch(
      channel_videos_query,
      playlistId=videos_playlist.to_list(row_type='scalar', distinct=True),
      maxResults=50,
    ).to_list(row_type='scalar', distinct=True)

    video_performance_query = """
    SELECT
      id AS media_url,
      snippet.title AS media_name,
      contentDetails.duration AS video_duration,
      statistics.viewCount AS views,
      statistics.likeCount AS likes
    FROM videos
    """
    return youtube_api_fetcher.fetch(video_performance_query, id=videos)



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=C0330, g-bad-import-order, g-multiple-import

"""Responsible for fetching media specific information from various sources."""

from __future__ import annotations

import enum
import logging
import warnings
from typing import Any, get_args

from garf_core import report

from media_fetching import exceptions
from media_fetching.enrichers import enricher
from media_fetching.sources import fetcher, models

logger = logging.getLogger('media-fetching')


class MediaFetchingService:
  """Extracts media information from a specified source.

  Attributes:
    source_fetcher: Initialized concrete fetcher for getting media data.
    source: Alias of a source.
  """

  def __init__(
    self,
    source: str | models.InputSource | None = None,
    source_fetcher: models.BaseMediaInfoFetcher | None = None,
  ) -> None:
    """Initializes MediaFetchingService.

    Args:
      source: [Deprecated] Alias of a source.
      source_fetcher: Initialized concrete fetcher for getting media data.
    """
    if not source and not source_fetcher:
      raise exceptions.MediaFetchingError('Missing source_fetcher parameter.')
    if source and not source_fetcher:
      warnings.warn(
        'Initialization from source argument is deprecated.'
        'Provide initialized source_fetcher instead',
        FutureWarning,
        stacklevel=2,
      )
      self.fetcher = _get_source_fetcher(source)
      self.source = source
    else:
      fetcher_alias = None
      for alias, (_, fetcher_class) in fetcher.FETCHERS.items():
        if isinstance(source_fetcher, fetcher_class):
          fetcher_alias = alias
          break

      if not fetcher_alias:
        raise exceptions.MediaFetchingError(
          f'Unsupported fetcher: {source_fetcher.__class__.__name__}'
        )
      self.fetcher = source_fetcher
      self.source = fetcher_alias

  @classmethod
  def from_source_alias(
    cls,
    source: str | models.InputSource = 'googleads',
  ) -> MediaFetchingService:
    """Initialized MediaFetchingService from source alias."""
    source_fetcher = _get_source_fetcher(source)
    return cls(source_fetcher=source_fetcher)

  def fetch(
    self,
    request: models.FetchingParameters,
    extra_parameters: dict[str, dict[str, Any]] | None = None,
  ) -> report.GarfReport:
    """Extracts data from specified source."""
    logger.info(
      "Fetching data from source '%s' with parameters: %s",
      self.source,
      request,
    )
    media_data = self.fetcher.fetch_media_data(request)
    if extra_info_modules := request.extra_info:
      extra_data = enricher.prepare_extra_info(
        performance=media_data,
        media_type=request.media_type,
        modules=extra_info_modules,
        params=extra_parameters,
      )
      enricher.enrich(media_data, extra_data)
    return media_data


def _get_source_fetcher(
  source: str | models.InputSource,
) -> models.BaseMediaInfoFetcher:
  if isinstance(source, enum.Enum):
    source = source.value
  if not (fetcher_info := fetcher.FETCHERS.get(source)):
    raise exceptions.MediaFetchingError(
      f'Incorrect source: {source}. Only {get_args(models.InputSource)} '
      'are supported.'
    )
  return fetcher_info[1]()



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=C0330, g-bad-import-order, g-multiple-import

"""Defines fetching data from a fake datasource."""

from collections.abc import Sequence
from typing import Literal

import garf_core
import pydantic
from media_tagging import media

from media_fetching.sources import models


class FakeFetchingParameters(models.FetchingParameters):
  """Parameters for getting media from a supplied data."""

  media_type: Literal[tuple(media.MediaTypeEnum.options())] | None = None
  media_identifier: str = 'media_url'
  media_name: str = 'media_name'
  metrics: Sequence[str] | str = ('clicks', 'impressions')
  segments: Sequence[str] | None = None
  extra_info: Sequence[str] | None = pydantic.Field(default_factory=list)

  def model_post_init(self, __context__):
    if isinstance(self.metrics, str):
      self.metrics = self.metrics.split(',')


class FakeFetcher(models.BaseMediaInfoFetcher):
  """Extracts information from a file to a media report."""

  def __init__(self, data: garf_core.GarfReport) -> None:
    self.data = data

  def fetch_media_data(
    self,
    fetching_request: FakeFetchingParameters = FakeFetchingParameters(),
  ) -> garf_core.GarfReport:
    return self.data



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=C0330, g-bad-import-order, g-multiple-import

"""Models and interfaces for media fetching library."""

import abc
import enum

import pydantic
from garf_core import report


class InputSource(str, enum.Enum):
  """Specifies supported sources for getting media performance."""

  googleads = 'googleads'
  youtube = 'youtube'
  file = 'file'
  bq = 'bq'
  sqldb = 'sqldb'
  fake = 'fake'
  dbm = 'dbm'


class FetchingParameters(pydantic.BaseModel):
  """Interface for parameters for getting media data."""

  model_config = pydantic.ConfigDict(extra='allow')


class BaseMediaInfoFetcher(abc.ABC):
  """Interface for getting data from a source."""

  @abc.abstractmethod
  def fetch_media_data(
    self,
    fetching_request: FetchingParameters,
  ) -> report.GarfReport:
    """Extracts data from a source as a report."""



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=C0330, g-bad-import-order, g-multiple-import

"""Enricher injects extra data into performance reports."""

from collections.abc import Sequence

from garf_core import report

from media_fetching import enrichers
from media_fetching.enrichers import extra_info

AVAILABLE_MODULES = {
  'tagging': enrichers.MediaTaggingEnricher,
  'googleads': enrichers.GoogleAdsEnricher,
  'youtube': enrichers.YouTubeEnricher,
}


def prepare_extra_info(
  performance: report.GarfReport,
  media_type: str,
  modules: Sequence[str],
  params: dict[str, dict[str, str]],
) -> list[extra_info.ExtraInfo]:
  """Builds extra info based on performance report and specified modules.

  Args:
    performance: Report with performance data.
    media_type:  Type of media in the report.
    modules: Modules used to perform enriching.
    params: Parameters to perform enriching.

  Returns:
    All extra info to be injected into the report.
  """
  data = []
  if isinstance(modules, str):
    modules = modules.split(',')
  for module in modules:
    enricher_module, method = module.split('.', maxsplit=2)
    if available_module := AVAILABLE_MODULES.get(enricher_module):
      initialized_module = available_module(**params.get(enricher_module, {}))
      info = getattr(initialized_module, method)(
        performance, media_type=media_type
      )
      data.append(info)
  return data


def enrich(
  performance_report: report.GarfReport,
  extra_data: Sequence[extra_info.ExtraInfo],
) -> None:
  """Adds additional information to existing performance report.

  Args:
    performance_report: Report with performance data.
    extra_data: Information to be injected into performance report.
  """
  for row in performance_report:
    for data in extra_data:
      if not data.info:
        continue
      columns = list(data.info.values())[0].keys()
      for column in columns:
        search_key = row[data.base_key]
        row[column] = data.info.get(search_key, {}).get(column)



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=C0330, g-bad-import-order, g-multiple-import

"""CLI entrypoint for fetching media data."""

from typing import Optional

import typer
from garf_executors.entrypoints import utils as garf_utils
from garf_io import writer as garf_writer
from media_tagging import media
from media_tagging.entrypoints import utils as tagging_utils
from typing_extensions import Annotated

import media_fetching
from media_fetching.sources import fetcher, models

typer_app = typer.Typer()


def _version_callback(show_version: bool) -> None:
  if show_version:
    print(f'media-fetcher version: {media_fetching.__version__}')
    raise typer.Exit()


@typer_app.command(
  context_settings={'allow_extra_args': True, 'ignore_unknown_options': True}
)
@tagging_utils.log_shutdown
def main(
  ctx: typer.Context,
  media_type: Annotated[
    media.MediaTypeEnum,
    typer.Option(
      help='Type of media',
      case_sensitive=False,
    ),
  ] = 'IMAGE',
  source: Annotated[
    models.InputSource,
    typer.Option(
      help='Type of media',
      case_sensitive=False,
    ),
  ] = 'googleads',
  extra_info: Annotated[
    Optional[str],
    typer.Option(
      help=(
        'Comma separated modules to add extra information to fetched data '
        'specified in "module.method" format'
      )
    ),
  ] = None,
  writer: Annotated[
    garf_writer.WriterOption,
    typer.Option(
      help='Type of writer used to write resulting report',
    ),
  ] = 'json',
  output: Annotated[
    str,
    typer.Option(
      help='Destination of written report',
    ),
  ] = 'media_results',
  logger: Annotated[
    garf_utils.LoggerEnum,
    typer.Option(
      help='Type of logger',
    ),
  ] = 'rich',
  loglevel: Annotated[
    str,
    typer.Option(
      help='Level of logging',
    ),
  ] = 'INFO',
  log_name: Annotated[
    str,
    typer.Option(
      help='Name of logger',
    ),
  ] = 'media-fetcher',
  version: Annotated[
    bool,
    typer.Option(
      help='Display library version',
      callback=_version_callback,
      is_eager=True,
      expose_value=False,
    ),
  ] = False,
):  # noqa: D103
  garf_utils.init_logging(
    loglevel=loglevel.upper(), logger_type=logger, name=log_name
  )

  supported_enrichers = (
    media_fetching.enrichers.enricher.AVAILABLE_MODULES.keys()
  )
  parsed_param_keys = set([source, writer] + list(supported_enrichers))
  extra_parameters = garf_utils.ParamsParser(parsed_param_keys).parse(ctx.args)
  fetching_service = media_fetching.MediaFetchingService.from_source_alias(
    source
  )
  request_class, _ = fetcher.FETCHERS.get(source)
  request = request_class(
    **extra_parameters.get(source),
    extra_info=extra_info,
    media_type=media_type,
  )

  report = fetching_service.fetch(
    request=request,
    extra_parameters=extra_parameters,
  )
  garf_writer.create_writer(
    writer, **(extra_parameters.get(writer) or {})
  ).write(report, output)


if __name__ == '__main__':
  typer_app()



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=C0330, g-bad-import-order, g-multiple-import

"""Instantiates concrete fetcher and builds fetching context."""

from media_fetching.sources import dbm, fake, file, googleads, sql, youtube

FETCHERS = {
  'youtube': (youtube.YouTubeFetchingParameters, youtube.Fetcher),
  'googleads': (googleads.GoogleAdsFetchingParameters, googleads.Fetcher),
  'file': (file.FileFetchingParameters, file.Fetcher),
  'bq': (sql.BigQueryFetchingParameters, sql.BigQueryFetcher),
  'sqldb': (sql.SqlAlchemyQueryFetchingParameters, sql.SqlAlchemyQueryFetcher),
  'fake': (fake.FakeFetchingParameters, fake.FakeFetcher),
  'dbm': (dbm.BidManagerFetchingParameters, dbm.Fetcher),
}



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=C0330, g-bad-import-order, g-multiple-import

"""Contains Google Ads API queries."""

from __future__ import annotations

import dataclasses
from typing import Literal

from gaarf import base_query

SupportedMediaTypes = Literal['IMAGE', 'YOUTUBE_VIDEO', 'TEXT']
SupportedCampaignTypes = Literal[
  'pmax', 'app', 'demandgen', 'video', 'display', 'search'
]


class PerformanceQuery(base_query.BaseQuery):
  """Enforces presence of certain fields in the query.

  Attributes:
    base_query_text:
      A Gaarf query text template that contains aliases specified
      in `required_fields`.

  Raises:
    ValueError:
      If subclass query_text does not contain all required fields.
  """

  query_text = ''
  required_fields = (
    'date',
    'campaign_type',
    'channel_type',
    'format_type',
    'media_name',
    'media_url',
    'aspect_ratio',
    'clicks',
    'impressions',
    'cost',
    'conversions',
    'conversions_value',
  )

  def __init_subclass__(cls) -> None:  # noqa: D105
    super().__init_subclass__()
    missing_fields: list[str] = []
    missing_fields = [
      field for field in cls.required_fields if field not in cls.query_text
    ]
    if missing_fields:
      raise ValueError(
        f'query_text does not contain required fields: {missing_fields}'
      )


@dataclasses.dataclass
class DisplayAssetPerformance(PerformanceQuery):
  """Fetches image ads performance for Display campaigns."""

  query_text = """
  SELECT
    '{campaign_type}' AS campaign_type,
    segments.date AS date,
    campaign.id AS campaign_id,
    campaign.advertising_channel_type AS channel_type,
    ad_group_ad.ad.name AS media_name,
    ad_group_ad.ad.id AS asset_id,
    ad_group_ad.ad.image_ad.image_url AS media_url,
    'UNKNOWN' AS format_type,
    ad_group_ad.ad.image_ad.pixel_width / ad_group_ad.ad.image_ad.pixel_height
      AS aspect_ratio,
    0 AS file_size,
    ad_group_ad.policy_summary.approval_status AS approval_status,
    metrics.cost_micros / 1e6 AS cost,
    metrics.clicks AS clicks,
    metrics.impressions AS impressions,
    metrics.conversions AS conversions,
    metrics.conversions_value AS conversions_value
  FROM ad_group_ad
  WHERE
    ad_group_ad.ad.type = IMAGE_AD
    AND campaign.advertising_channel_type = DISPLAY
    AND segments.date BETWEEN '{start_date}' AND '{end_date}'
    AND ad_group_ad.ad.image_ad.image_url != ''
    AND metrics.cost_micros > {min_cost}
  """

  start_date: str
  end_date: str
  media_type: SupportedMediaTypes
  campaign_type: SupportedCampaignTypes
  min_cost: int = 0

  def __post_init__(self) -> None:  # noqa: D105
    self.min_cost = int(self.min_cost * 1e6)


@dataclasses.dataclass
class VideoPerformance(PerformanceQuery):
  """Fetches video ad performance for Video campaigns."""

  query_text = """
  SELECT
    '{campaign_type}' AS campaign_type,
    campaign.id AS campaign_id,
    segments.date AS date,
    campaign.advertising_channel_type AS channel_type,
    ad_group_ad.ad.type AS ad_type,
    video.id AS media_url,
    video.title AS media_name,
    segments.ad_format_type AS format_type,
    0 AS aspect_ratio,
    video.duration_millis / 1000 AS video_duration,
    ad_group_ad.policy_summary.approval_status AS approval_status,
    metrics.cost_micros / 1e6 AS cost,
    metrics.clicks AS clicks,
    metrics.impressions AS impressions,
    metrics.conversions AS conversions,
    metrics.conversions_value AS conversions_value
  FROM video
  WHERE
    campaign.advertising_channel_type = VIDEO
    AND segments.date BETWEEN '{start_date}' AND '{end_date}'
    AND video.id != ''
    AND metrics.cost_micros > {min_cost}
  """

  start_date: str
  end_date: str
  media_type: SupportedMediaTypes
  campaign_type: SupportedCampaignTypes
  min_cost: int = 0

  def __post_init__(self) -> None:  # noqa: D105
    self.min_cost = int(self.min_cost * 1e6)


@dataclasses.dataclass
class PmaxAssetPerformance(PerformanceQuery):
  """Fetches asset info for Performance Max campaigns."""

  query_text = """
  SELECT
    '{campaign_type}' AS campaign_type,
    segments.date AS date,
    campaign.id AS campaign_id,
    campaign.advertising_channel_type AS channel_type,
    asset.id AS asset_id,
    {media_name} AS media_name,
    {media_url} AS media_url,
    asset_group_asset.field_type AS format_type,
    {aspect_ratio} AS aspect_ratio,
    {size} AS {size_column},
    asset_group_asset.policy_summary.approval_status AS approval_status,
    0 AS cost,
    0 AS clicks,
    0 AS impressions,
    metrics.conversions AS conversions,
    metrics.conversions_value AS conversions_value
  FROM asset_group_asset
  WHERE
    asset.type = {media_type}
    AND campaign.advertising_channel_type = PERFORMANCE_MAX
    AND {media_url} != ''
    AND segments.date BETWEEN '{start_date}' AND '{end_date}'
  """

  start_date: str
  end_date: str
  media_type: SupportedMediaTypes
  campaign_type: SupportedCampaignTypes
  min_cost: int = 0

  def __post_init__(self) -> None:  # noqa: D105
    if self.media_type == 'IMAGE':
      self.media_url = 'asset.image_asset.full_size.url'
      self.aspect_ratio = (
        'asset.image_asset.full_size.width_pixels / '
        'asset.image_asset.full_size.height_pixels'
      )
      self.size = 'asset.image_asset.file_size / 1024'
      self.size_column = 'file_size'
      self.media_name = 'asset.name'
    elif self.media_type == 'TEXT':
      self.media_url = 'asset.text_asset.text'
      self.aspect_ratio = '0'
      self.size = '0'
      self.size_column = 'file_size'
      self.media_name = 'asset.text_asset.text'
    else:
      self.media_url = 'asset.youtube_video_asset.youtube_video_id'
      self.aspect_ratio = 0.0
      self.size = 0.0
      self.size_column = 'video_duration'
      self.media_name = 'asset.youtube_video_asset.youtube_video_title'


@dataclasses.dataclass
class SearchAssetPerformance(PerformanceQuery):
  """Fetches text asset performance for Search campaigns."""

  query_text = """
  SELECT
    '{campaign_type}' AS campaign_type,
    segments.date AS date,
    campaign.id AS campaign_id,
    campaign.advertising_channel_type AS channel_type,
    asset.text_asset.text AS media_name,
    asset.id AS asset_id,
    asset.text_asset.text AS media_url,
    ad_group_ad_asset_view.field_type AS format_type,
    0 AS aspect_ratio,
    0 AS file_size,
    ad_group_ad.ad.name AS ad_name,
    ad_group_ad_asset_view.policy_summary:approval_status AS approval_status,
    metrics.cost_micros / 1e6 AS cost,
    metrics.clicks AS clicks,
    metrics.impressions AS impressions,
    metrics.conversions AS conversions,
    metrics.conversions_value AS conversions_value
  FROM ad_group_ad_asset_view
  WHERE
    campaign.advertising_channel_type =  SEARCH
    AND asset.type = TEXT
    AND asset.text_asset.text != ''
    AND segments.date BETWEEN '{start_date}' AND '{end_date}'
    AND metrics.cost_micros > {min_cost}
  """

  start_date: str
  end_date: str
  media_type: SupportedMediaTypes
  campaign_type: SupportedCampaignTypes
  min_cost: int = 0

  def __post_init__(self) -> None:  # noqa: D105
    self.min_cost = int(self.min_cost * 1e6)


@dataclasses.dataclass
class DemandGenTextAssetPerformance(PerformanceQuery):
  """Fetches image asset performance for Demand Gen campaigns."""

  query_text = """
  SELECT
    '{campaign_type}' AS campaign_type,
    segments.date AS date,
    campaign.id AS campaign_id,
    campaign.advertising_channel_type AS channel_type,
    asset.text_asset.text AS media_name,
    asset.id AS asset_id,
    asset.text_asset.text AS media_url,
    ad_group_ad_asset_view.field_type AS format_type,
    0 AS aspect_ratio,
    0 AS file_size,
    ad_group_ad.ad.name AS ad_name,
    ad_group_ad_asset_view.policy_summary:approval_status AS approval_status,
    metrics.cost_micros / 1e6 AS cost,
    metrics.clicks AS clicks,
    metrics.impressions AS impressions,
    metrics.conversions AS conversions,
    metrics.conversions_value AS conversions_value
  FROM ad_group_ad_asset_view
  WHERE
    campaign.advertising_channel_type =  DEMAND_GEN
    AND asset.type = TEXT
    AND asset.text_asset.text != ''
    AND segments.date BETWEEN '{start_date}' AND '{end_date}'
    AND metrics.cost_micros > {min_cost}
  """

  start_date: str
  end_date: str
  media_type: SupportedMediaTypes
  campaign_type: SupportedCampaignTypes
  min_cost: int = 0

  def __post_init__(self) -> None:  # noqa: D105
    self.min_cost = int(self.min_cost * 1e6)


@dataclasses.dataclass
class DemandGenImageAssetPerformance(PerformanceQuery):
  """Fetches image asset performance for Demand Gen campaigns."""

  query_text = """
  SELECT
    '{campaign_type}' AS campaign_type,
    segments.date AS date,
    campaign.id AS campaign_id,
    campaign.advertising_channel_type AS channel_type,
    asset.name AS media_name,
    asset.id AS asset_id,
    asset.image_asset.full_size.url AS media_url,
    ad_group_ad_asset_view.field_type AS format_type,
    asset.image_asset.full_size.width_pixels /
      asset.image_asset.full_size.height_pixels AS aspect_ratio,
    asset.image_asset.file_size / 1024 AS file_size,
    ad_group_ad.ad.name AS ad_name,
    ad_group_ad_asset_view.policy_summary:approval_status AS approval_status,
    metrics.cost_micros / 1e6 AS cost,
    metrics.clicks AS clicks,
    metrics.impressions AS impressions,
    metrics.conversions AS conversions,
    metrics.conversions_value AS conversions_value
  FROM ad_group_ad_asset_view
  WHERE
    campaign.advertising_channel_type =  DEMAND_GEN
    AND asset.type = IMAGE
    AND ad_group_ad_asset_view.field_type NOT IN (
      BUSINESS_LOGO, LANDSCAPE_LOGO,  LOGO
    )
    AND segments.date BETWEEN '{start_date}' AND '{end_date}'
    AND asset.image_asset.full_size.url != ''
    AND metrics.cost_micros > {min_cost}
  """

  start_date: str
  end_date: str
  media_type: SupportedMediaTypes
  campaign_type: SupportedCampaignTypes
  min_cost: int = 0

  def __post_init__(self) -> None:  # noqa: D105
    self.min_cost = int(self.min_cost * 1e6)


@dataclasses.dataclass
class DemandGenVideoAssetPerformance(PerformanceQuery):
  """Fetches video asset performance for Demand Gen campaigns."""

  query_text = """
  SELECT
    '{campaign_type}' AS campaign_type,
    segments.date AS date,
    campaign.id AS campaign_id,
    campaign.advertising_channel_type AS channel_type,
    video.id AS media_url,
    video.title AS media_name,
    segments.ad_format_type AS format_type,
    0 AS aspect_ratio,
    video.duration_millis / 1000 AS video_duration,
    ad_group_ad.policy_summary.approval_status AS approval_status,
    metrics.cost_micros / 1e6 AS cost,
    metrics.clicks AS clicks,
    metrics.impressions AS impressions,
    metrics.conversions AS conversions,
    metrics.conversions_value AS conversions_value
  FROM video
  WHERE
    campaign.advertising_channel_type =  DEMAND_GEN
    AND segments.date BETWEEN '{start_date}' AND '{end_date}'
    AND video.id != ''
    AND metrics.cost_micros > {min_cost}
  """

  start_date: str
  end_date: str
  media_type: SupportedMediaTypes
  campaign_type: SupportedCampaignTypes
  min_cost: int = 0

  def __post_init__(self) -> None:  # noqa: D105
    self.min_cost = int(self.min_cost * 1e6)


@dataclasses.dataclass
class AppAssetPerformance(PerformanceQuery):
  """Fetches asset performance for App campaigns."""

  query_text = """
  SELECT
    '{campaign_type}' AS campaign_type,
    segments.date AS date,
    campaign.id AS campaign_id,
    asset.id AS asset_id,
    {media_name} AS media_name,
    {media_url} AS media_url,
    ad_group_ad_asset_view.field_type AS format_type,
    {aspect_ratio} AS aspect_ratio,
    {size} AS {size_column},
    ad_group_ad_asset_view.policy_summary:approval_status AS approval_status,
    metrics.cost_micros / 1e6 AS cost,
    metrics.clicks AS clicks,
    metrics.impressions AS impressions,
    metrics.conversions AS conversions,
    metrics.biddable_app_install_conversions AS installs,
    metrics.biddable_app_post_install_conversions AS inapps,
    metrics.conversions_value AS conversions_value
  FROM ad_group_ad_asset_view
  WHERE
    campaign.advertising_channel_type = MULTI_CHANNEL
    AND asset.type = {media_type}
    AND segments.date BETWEEN '{start_date}' AND '{end_date}'
    AND {media_url} != ''
    AND metrics.cost_micros > {min_cost}
    {app_id}
  """

  start_date: str
  end_date: str
  media_type: SupportedMediaTypes
  campaign_type: SupportedCampaignTypes
  min_cost: int = 0
  app_id: str | None = None

  def __post_init__(self) -> None:  # noqa: D105
    self.app_id = (
      f'AND campaign.app_campaign_setting.app_id = "{self.app_id}"'
      if self.app_id
      else ''
    )
    self.min_cost = int(self.min_cost * 1e6)
    if self.media_type == 'IMAGE':
      self.media_url = 'asset.image_asset.full_size.url'
      self.aspect_ratio = (
        'asset.image_asset.full_size.width_pixels / '
        'asset.image_asset.full_size.height_pixels'
      )
      self.size = 'asset.image_asset.file_size / 1024'
      self.size_column = 'file_size'
      self.media_name = 'asset.name'
    elif self.media_type == 'TEXT':
      self.media_url = 'asset.text_asset.text'
      self.aspect_ratio = '0'
      self.size = '0'
      self.size_column = 'file_size'
      self.media_name = 'asset.text_asset.text'
    else:
      self.media_url = 'asset.youtube_video_asset.youtube_video_id'
      self.aspect_ratio = 0.0
      self.size = 0.0
      self.size_column = 'video_duration'
      self.media_name = 'asset.youtube_video_asset.youtube_video_title'


QUERIES_MAPPING: dict[
  str, base_query.BaseQuery | dict[str, base_query.BaseQuery]
] = {
  'search': SearchAssetPerformance,
  'app': AppAssetPerformance,
  'display': DisplayAssetPerformance,
  'pmax': PmaxAssetPerformance,
  'video': VideoPerformance,
  'demandgen': {
    'YOUTUBE_VIDEO': DemandGenVideoAssetPerformance,
    'IMAGE': DemandGenImageAssetPerformance,
    'TEXT': DemandGenTextAssetPerformance,
  },
}

CAMPAIGN_TYPES_MAPPING: dict[str, str] = {
  'app': 'MULTI_CHANNEL',
  'search': 'SEARCH',
  'display': 'DISPLAY',
  'pmax': 'PERFORMANCE_MAX',
  'video': 'VIDEO',
  'demandgen': 'DEMAND_GEN',
}
